/*
 * Copyright 2018-2021 Crown Copyright
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package uk.gov.gchq.gaffer.parquetstore.utils;

import com.google.common.collect.Sets;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.parquet.hadoop.ParquetWriter;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;
import scala.collection.JavaConversions$;
import scala.collection.mutable.WrappedArray;

import uk.gov.gchq.gaffer.commonutil.TestGroups;
import uk.gov.gchq.gaffer.data.element.Element;
import uk.gov.gchq.gaffer.parquetstore.ParquetStore;
import uk.gov.gchq.gaffer.parquetstore.io.writer.ParquetElementWriter;
import uk.gov.gchq.gaffer.parquetstore.operation.handler.utilities.AggregateDataForGroup;
import uk.gov.gchq.gaffer.parquetstore.testutils.DataGen;
import uk.gov.gchq.gaffer.parquetstore.testutils.TestUtils;
import uk.gov.gchq.gaffer.spark.SparkSessionProvider;
import uk.gov.gchq.gaffer.types.FreqMap;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

public class AggregateDataForGroupTest {

    private static void generateData(final String file, final SchemaUtils schemaUtils) throws IOException {
        final ParquetWriter<Element> writer = new ParquetElementWriter
                .Builder(new Path(file))
                .withSparkSchema(schemaUtils.getSparkSchema(TestGroups.ENTITY))
                .withType(schemaUtils.getParquetSchema(TestGroups.ENTITY))
                .usingConverter(schemaUtils.getConverter(TestGroups.ENTITY))
                .build();
        for (int i = 19; i >= 0; i--) {
            writer.write(DataGen.getEntity(TestGroups.ENTITY, (long) i, (byte) 'a', 3f,
                    TestUtils.getTreeSet1(), 5L * i, (short) 6, TestUtils.DATE, TestUtils.getFreqMap1(), 1, "A"));
            writer.write(DataGen.getEntity(TestGroups.ENTITY, (long) i, (byte) 'b', 4f,
                    TestUtils.getTreeSet2(), 6L * i, (short) 7, TestUtils.DATE, TestUtils.getFreqMap2(), 1, "A"));
        }
        writer.close();
    }

    @Test
    public void aggregateDataForGroupTest(@TempDir java.nio.file.Path tempDir)
            throws Exception {
        // Given
        final SchemaUtils schemaUtils = new SchemaUtils(TestUtils.gafferSchema("schemaUsingLongVertexType"));
        final String file1 = tempDir.resolve("inputdata1.parquet").toString();
        final String file2 = tempDir.resolve("inputdata2.parquet").toString();
        generateData(file1, schemaUtils);
        generateData(file2, schemaUtils);
        final SparkSession sparkSession = SparkSessionProvider.getSparkSession();
        final List<String> inputFiles = new ArrayList<>(Sets.newHashSet(file1, file2));
        final String outputFolder = tempDir.resolve("aggregated").toString();
        final AggregateDataForGroup aggregator = new AggregateDataForGroup(FileSystem.get(new Configuration()), schemaUtils, TestGroups.ENTITY,
                inputFiles, outputFolder, sparkSession);

        // When
        aggregator.call();

        // Then
        final FileSystem fs = FileSystem.get(new Configuration());
        assertTrue(fs.exists(new Path(outputFolder)));
        final Row[] results = (Row[]) sparkSession
                .read()
                .parquet(outputFolder)
                .sort(ParquetStore.VERTEX)
                .collect();
        for (int i = 0; i < 20; i++) {
            assertEquals((long) i, (long) results[i].getAs(ParquetStore.VERTEX));
            assertEquals('b', ((byte[]) results[i].getAs("byte"))[0]);
            assertEquals(14f, results[i].getAs("float"), 0.01f);
            assertEquals(11L * 2 * i, (long) results[i].getAs("long"));
            assertEquals(26, (int) results[i].getAs("short"));
            assertEquals(TestUtils.DATE.getTime(), (long) results[i].getAs("date"));
            assertEquals(4, (int) results[i].getAs("count"));
            assertArrayEquals(new String[]{"A", "B", "C"},
                    (String[]) ((WrappedArray<String>) results[i].getAs("treeSet")).array());
            final FreqMap mergedFreqMap = new FreqMap();
            mergedFreqMap.put("A", 4L);
            mergedFreqMap.put("B", 2L);
            mergedFreqMap.put("C", 2L);
            assertEquals(JavaConversions$.MODULE$.mapAsScalaMap(mergedFreqMap), results[i].getAs("freqMap"));
        }
    }
}
